/*
 * $RCSfile: SharedMemory.cc,v $
 * $Revision: 1.4 $
 *
 * Comments:
 *
 * (C) Copyright ParaSoft Corporation 1996.  All rights reserved.
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft
 * The copyright notice above does not evidence any
 * actual or intended publication of such source code.
 *
 * $Author: dario $          $Locker:  $
 * $Date: 2007-04-23 12:17:40 $
 * $Log: not supported by cvs2svn $
 * Revision 1.3  2007/02/02 15:43:50  dario
 * compile error fixed
 *
 * Revision 1.2  2007/02/02 15:25:35  mali
 * error handling added
 *
 * Revision 1.1  2005/08/09 07:51:55  mlyko
 * Original source code.
 *
 *
 */

#include <stdio.h>
#include <memory.h>

#include "SharedMemory.h"

/** Shared memory region size */
const static int SM_ALLOC_SIZE = 4096;

/** String object identifier (java) */
const static char * JAVA_STRING_IDENTIFIER = "Ljava/lang/String;";

/** Key variable name */
const static char * JAVA_KEY_VARIABLE_NAME = "_sKey";

static ErrorDescription __error;

SharedMemory::SharedMemory() : _smHandle(0)
{
}

SharedMemory::~SharedMemory()
{
    //printf("Destroy!\n");
}

ErrorDescription& SharedMemory::getErrorDescription()
{
    return __error;
}

bool SharedMemory::createHandle(JNIEnv * env, jobject obj)
{
    jstring jKeyName = askJavaForKeyName(env, obj);
    jboolean isCopy;
    const char* smKey = env->GetStringUTFChars(jKeyName, &isCopy);
    _smHandle = createMap(smKey);
    if (isCopy == JNI_TRUE) {
        env->ReleaseStringUTFChars(jKeyName, smKey);
    }
    if (_smHandle) {
        return true;
    }
    return false;
}

bool SharedMemory::getHandle(JNIEnv * env, jobject obj)
{
    jstring jKeyName = askJavaForKeyName(env, obj);
    jboolean isCopy;
    const char* smKey = env->GetStringUTFChars(jKeyName, &isCopy);
    HANDLE smHandle = openMap(smKey);
    if (isCopy == JNI_TRUE) {
        env->ReleaseStringUTFChars(jKeyName, smKey);
    }
    if (smHandle) {
        if (_smHandle) {
            /* Keep only one handle to shared memory */
            closeHandle(smHandle);
        } else {
            _smHandle = smHandle;
        }
        return true;
    }
    return false;
}

jstring SharedMemory::askJavaForKeyName(JNIEnv * env, jobject obj)
{
    jstring jKeyName;
    jclass cls = env->GetObjectClass(obj);
    jfieldID fid = env->GetFieldID(cls, JAVA_KEY_VARIABLE_NAME, JAVA_STRING_IDENTIFIER);
    if (fid == 0) {
        printf("Failed to get java variable field id.\n");
        return env->NewStringUTF("");
    } else {
        jKeyName = (jstring)env->GetObjectField(obj, fid);
    }
    return jKeyName;
}

bool SharedMemory::setByteArray(JNIEnv * env, jobject obj, jbyteArray byteArray)
{
    void * pSharedMemory = (void *) mapView(_smHandle);
    if (!pSharedMemory) {
        printf("Can't create pointer to shared memory\n");
        return false;
    }
    jsize len = env->GetArrayLength(byteArray);
    if (len > SM_ALLOC_SIZE) {
        len = SM_ALLOC_SIZE;
    }
    jbyte * byteArrayBody = env->GetByteArrayElements(byteArray, 0);
    memcpy(pSharedMemory, byteArrayBody, len);
    env->ReleaseByteArrayElements(byteArray, byteArrayBody, 0);
    unmapView(pSharedMemory);
    return true;
}

bool SharedMemory::getByteArray(JNIEnv * env, jobject obj, jbyteArray byteArray)
{
    int len = env->GetArrayLength(byteArray);
    if (len > SM_ALLOC_SIZE) {
        printf("Size out of range.");
        return false;
    }
    jbyte * pSharedMemory = (jbyte *) mapView(_smHandle);
    if (!pSharedMemory) {
        printf("Can't create pointer to shared memory\n");
        return false;
    }
    env->SetByteArrayRegion(byteArray, 0, len, pSharedMemory);
    unmapView(pSharedMemory);
    return true;
}

HANDLE SharedMemory::createMap(LPCTSTR MapName)
{
    __error.clearError();
    HANDLE hMap = CreateFileMapping(INVALID_HANDLE_VALUE,
                                0,
                                PAGE_READWRITE,
                                0,
                                SM_ALLOC_SIZE,
                                MapName);

    if (hMap == NULL || hMap == INVALID_HANDLE_VALUE) {
        printf("Failed to create shared memory mapping object: %d\n", GetLastError());
        __error.setError(GetLastError(), "CreateFileMapping", "Failed to create shared memory mapping object!");
        return NULL;
    }
    return hMap;
}

HANDLE SharedMemory::openMap(LPCTSTR mapName)
{
    HANDLE hMap;
    __error.clearError();
    hMap = OpenFileMapping(FILE_MAP_READ | FILE_MAP_WRITE,
                          TRUE,
                          mapName);
                          
    if (hMap == NULL || hMap == INVALID_HANDLE_VALUE) {
        __error.setError(GetLastError(), "OpenFileMapping", "Failed to open shared memory mapping object!");
        return NULL;
    }
    
    return hMap;
}

LPVOID SharedMemory::mapView(HANDLE hMap)
{
    LPVOID MappedPointer;

    __error.clearError();
    MappedPointer= MapViewOfFile(hMap,
                                FILE_MAP_WRITE | FILE_MAP_READ,
                                0, 0, SM_ALLOC_SIZE);
    if (MappedPointer != NULL) {
        return MappedPointer;
    }
    __error.setError(GetLastError(), "MapViewOfFile");
    return 0;
}

void SharedMemory::unmapView(LPVOID mapedView)
{
    __error.clearError();
    if (mapedView) {
        BOOL bResult = UnmapViewOfFile(mapedView);
        if (!bResult) {
            __error.setError(GetLastError(), "UnmapViewOfFile");
        }
        mapedView = 0;
    }
}

bool SharedMemory::unlockSharedMemory()
{
    closeHandle(_smHandle);
    _smHandle = NULL;
    return true;
}

void SharedMemory::closeHandle(HANDLE handle)
{
    __error.clearError();
    if (handle) {
        bool bResult = CloseHandle(handle);
        if (!bResult) {
            __error.setError(GetLastError(), "UnmapViewOfFile");
        }
        handle = 0;
    }
}
